// mSock.cpp (PSP Version)

#include "mSock.h"

#include <pspsdk.h>
#include <psputility.h>
#include <pspnet.h>
#include <pspnet_inet.h>
#include <pspnet_apctl.h>
#include <pspnet_resolver.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <errno.h>
#include <string>
#include <cstring> // for strlen()
using namespace std;

typedef int SOCKET;
#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
#define LASTERROR sceNetInetGetErrno()
#define SD_BOTH 2

// Configuration
#define MSOCK_MAX_BUFFER_SIZE 1024

//=========================================================
// Static Resolver
//=========================================================

static bool resolver_created;
static char resolverBuffer[1024];
static int resolverId = 0;

void resolver_create() {
	if (resolver_created) return;
	sceNetResolverCreate(&resolverId, resolverBuffer, sizeof(resolverBuffer));
	resolver_created = true;
}

//=========================================================
// Helpers
//=========================================================

struct mSock_data {
	SOCKET sock;
	bool bValid;
	string sCache;
	
	static void unblock_socket(SOCKET sock) {
		int iMode = 1;
		sceNetInetSetsockopt(sock, SOL_SOCKET, 0x1009, (const char*)&iMode, sizeof(unsigned int));
	};

	static void invalidate_data(mSock_data* d) {
		if (! d->bValid) return;
		shutdown(d->sock, SD_BOTH);
		sceNetInetClose(d->sock);
		d->sock = INVALID_SOCKET;
		d->bValid = false;
	};

	static void check_err(mSock_data* d) {
		int err = LASTERROR;
		if (err != 0 && err != EWOULDBLOCK) mSock_data::invalidate_data(d);
	};
};

//=========================================================
// Client socket
//=========================================================

mSock::mSock(const char* host, unsigned short port) {
	resolver_create();

	d = new mSock_data;
	d->bValid = false;
	
	sockaddr_in sa;
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	int err = sceNetInetInetAton(host, &sa.sin_addr);
	if (err == 0) {
		err = sceNetResolverStartNtoA(resolverId, host, &sa.sin_addr, 2, 3);
		if (err < 0) return;
	}
	
	d->sock = sceNetInetSocket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (d->sock & 0x80000000) return;
	
	if (sceNetInetConnect(d->sock, (sockaddr*)&sa, sizeof(sockaddr_in))) return;
	
	mSock_data::unblock_socket(d->sock);
	
	d->bValid = true;
	doEvents();
}

mSock::~mSock() {
	if (d->bValid) mSock_data::invalidate_data(d);
	delete d;
}

// Public

bool mSock::isValid() {
	return d->bValid;
}

void mSock::doEvents() {
	if (! d->bValid) return;
	char cBuff[MSOCK_MAX_BUFFER_SIZE] = {0};

	int recvVal = sceNetInetRecv(d->sock, cBuff, MSOCK_MAX_BUFFER_SIZE, 0);
	if (recvVal == SOCKET_ERROR) {
		mSock_data::check_err(d);
	} else if (recvVal == 0) {
		mSock_data::invalidate_data(d);
	} else {
		d->sCache.append(cBuff, recvVal); 
	}
}

string mSock::get() {
	string sResult;
	sResult.assign(d->sCache);
	d->sCache.assign("");
	return sResult;
}

void mSock::put(const char* data, int count) {
	if (! d->bValid) return;
	if (sceNetInetSend(d->sock, data, count, 0) == SOCKET_ERROR) {
		mSock_data::check_err(d);
	}
}

void mSock::put(const char *data) {
	this->put(data, (int)strlen(data) );
}

void mSock::close() {
	mSock_data::invalidate_data(d);
}